/*
 *   SlimProtoLib Copyright (c) 2004,2006 Richard Titmuss
 *
 *   This file is part of SlimProtoLib.
 *
 *   SlimProtoLib is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   SlimProtoLib is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SlimProtoLib; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include <jni.h>

#include "slimproto/slimproto.h"


typedef struct {
	slimproto_t p;
	JavaVM *vm;
	JNIEnv *env;
	jobject obj;
} jni_data_t;


static int connect_callback(slimproto_t *p, bool isConnected, void *user_data);
static int command_callback(slimproto_t *p, const unsigned char *buf, int buf_len, void *user_data);


JNIEXPORT jint JNICALL Java_org_titmuss_slimprotolib_SlimProto__1init
  (JNIEnv *env, jobject obj) {

	jni_data_t *j = (jni_data_t *) malloc(sizeof(jni_data_t));
	memset(j, 0, sizeof(jni_data_t));
	
	j->obj = (*env)->NewGlobalRef(env, obj);
  	slimproto_init(&j->p);  	
  	slimproto_add_connect_callback(&j->p, connect_callback, j);

  	return (jint) j;
}

JNIEXPORT void JNICALL Java_org_titmuss_slimprotolib_SlimProto__1close
  (JNIEnv *env, jobject obj, jint ptr) {
  	
	jni_data_t *j = (jni_data_t *) ptr;
  	slimproto_close(&j->p);
}

JNIEXPORT void JNICALL Java_org_titmuss_slimprotolib_SlimProto__1destroy
  (JNIEnv *env, jobject obj, jint ptr) {
  	
	jni_data_t *j = (jni_data_t *) ptr;
	// fixme remove connect callback
	// fixme remove command callbacks
  	slimproto_destroy(&j->p);
  	(*env)->DeleteGlobalRef(env, j->obj);
  	free(j);
}

JNIEXPORT jint JNICALL Java_org_titmuss_slimprotolib_SlimProto__1connect
  (JNIEnv *env, jobject obj, jint ptr, jstring jServerAddress, jint serverPort) {

	jni_data_t *j = (jni_data_t *) ptr;

    const char *serverAddress = (*env)->GetStringUTFChars(env, jServerAddress, NULL);  	
  	int r = slimproto_connect(&j->p, serverAddress, serverPort);
  	(*env)->ReleaseStringUTFChars(env, jServerAddress, serverAddress);
  	
  	return r;
}

JNIEXPORT jint JNICALL Java_org_titmuss_slimprotolib_SlimProto__1helo
  (JNIEnv *env, jobject obj, jint ptr, jint deviceId, jint revision, jbyteArray jMacAddress, jboolean isGraphics, jboolean isConnect) {
  	
	jni_data_t *j = (jni_data_t *) ptr;

	jbyte macAddress[6];	
	(*env)->GetByteArrayRegion(env, jMacAddress, 0, 6, macAddress);  	
  	int r = slimproto_helo(&j->p, deviceId, revision, (char *)macAddress, isGraphics, isConnect);

	return r;  	
}

JNIEXPORT jint JNICALL Java_org_titmuss_slimprotolib_SlimProto__1command
  (JNIEnv *env, jobject obj, jint ptr, jbyteArray jBuf, jint jiffiesIdx) {
  	
	jni_data_t *j = (jni_data_t *) ptr;
  	
 	jbyte *buf = (jbyte *) (*env)->GetIntArrayElements(env, jBuf, 0);
 	
 	if (jiffiesIdx > 0) {
 		slimproto_set_jiffies(&j->p, buf, jiffiesIdx);
 	} 	
  	return slimproto_send(&j->p, (char *)buf);
}
  
JNIEXPORT void JNICALL Java_org_titmuss_slimprotolib_SlimProto__1addCommandCallback
  (JNIEnv *env, jobject obj, jint ptr, jstring jCommand) {

	jni_data_t *j = (jni_data_t *) ptr;

    const char *command = (*env)->GetStringUTFChars(env, jCommand, NULL);  	
  	slimproto_add_command_callback(&j->p, command, command_callback, j);
  	(*env)->ReleaseStringUTFChars(env, jCommand, command);  	
}


static int connect_callback(slimproto_t *p, bool isConnected, void *user_data) {
	int r;
	
	jni_data_t *j = (jni_data_t *) user_data;
	
	if (isConnected) {
		jsize nVMs;
	
		r = JNI_GetCreatedJavaVMs(&j->vm, 1, &nVMs);
		if (r < 0) {
			fprintf(stderr, "Error in GetCreatedJavaVMs: %i\n", r);
			return -1;
		}
	
		(*j->vm)->AttachCurrentThread(j->vm, (void **)&j->env, NULL);
		if (r < 0) {
			fprintf(stderr, "Error in AttachCurrentThread: %i\n", r);
			return -1;
		}
	}
	
	JNIEnv *env = j->env;
	
	jclass clazz = (*env)->GetObjectClass(env, j->obj);	
	jmethodID method = (*env)->GetMethodID(env, clazz, "_connect_callback", "(Z)V");
	if (method == 0) {
		fprintf(stderr, "Cannot find SlimProto._connect_callback method\n");
		return -1;
	}
	
	jboolean jIsConnected = isConnected;
	
	(*env)->ExceptionClear(env);
	(*env)->CallVoidMethod(env, j->obj, method, jIsConnected);

	if((*env)->ExceptionOccurred(env)) {
		fprintf(stderr, "Exception in SlimProto._connect_callback method\n");
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		return -1;
    }
	
	if (!isConnected) {
		r = (*j->vm)->DetachCurrentThread(j->vm);
		if (r < 0) {
			fprintf(stderr, "Error in DetachCurrentThread: %i\n", r);
			return -1;
		}
		
		j->vm = NULL;
		j->env = NULL;	
	}

	return 0;
}

static int command_callback(slimproto_t *p, const unsigned char *buf, int buf_len, void *user_data) {
	jni_data_t *j = (jni_data_t *) user_data;

	JNIEnv *env = j->env;
	
	if ((*env)->PushLocalFrame(env, 5) < 0) {
		fprintf(stderr, "Cannot push local frame\n");
		return -1;		
	}
	
	jclass clazz = (*env)->GetObjectClass(env, j->obj);	
	jmethodID method = (*env)->GetMethodID(env, clazz, "_command_callback", "([B)V");
	if (method == 0) {
		fprintf(stderr, "Cannot find SlimProto._command_callback method\n");
		return -1;
	}
	
	jbyteArray jBuf = (*env)->NewByteArray(env, buf_len);
	if (jBuf == NULL) {
		fprintf(stderr, "Cannot create java byte[]\n");
		return -1;	
	}	
	(*env)->SetByteArrayRegion(env, jBuf, 0, buf_len, (jbyte *)buf);
		
	(*env)->ExceptionClear(env);
	(*env)->CallVoidMethod(env, j->obj, method, jBuf);

	if((*env)->ExceptionOccurred(env)) {
		fprintf(stderr, "Exception in SlimProto._command_callback method\n");
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		return -1;
    }

	(*env)->PopLocalFrame(env, NULL);	
	return 0;
}

