/*
 *   SlimProtoLib Copyright (c) 2004,2006 Richard Titmuss
 *
 *   This file is part of SlimProtoLib.
 *
 *   SlimProtoLib is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   SlimProtoLib is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SlimProtoLib; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>


#ifdef __WIN32__
  #include <winsock.h>
#else
  #include <arpa/inet.h>
  #include <unistd.h>
  #include <netinet/in.h>
  #include <netdb.h>
  #include <sys/socket.h>
  #include <sys/time.h>
  #include <string.h>
#endif


#include "slimproto/slimproto.h"
#include "slimaudio/slimaudio.h"


#ifdef SLIMPROTO_DEBUG
#define DEBUGF(...) if (slimaudio_debug) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUGF(...)
#endif


bool slimaudio_debug;


static int strm_callback(slimproto_t *p, const unsigned char *buf, int buf_len, void *user_data);
static void audio_stop(slimaudio_t *audio);


/*
 * Init the slimaudio layer:
 * - allocate the internal state
 * - register callbacks with slimproto
 */
int slimaudio_init(slimaudio_t *audio, slimproto_t *proto) {
	memset(audio, 0, sizeof(slimaudio_t));
	
	audio->proto = proto;
	audio->decoder_buffer = slimaudio_buffer_init(DECODER_BUFFER_SIZE);
	audio->output_buffer = slimaudio_buffer_init(OUTPUT_BUFFER_SIZE);
	
	if (slimaudio_output_init(audio) != 0)
		return -1;
	
	slimproto_add_command_callback(proto, "strm", &strm_callback, audio);
 	return 0;
}

/*
 * Frees slimaudio resources.
 */
void slimaudio_destroy(slimaudio_t *audio) {
	// FIXME remove slimproto callback		
	slimaudio_output_destroy(audio);	
	slimaudio_buffer_close(audio->output_buffer);
	slimaudio_buffer_close(audio->decoder_buffer);
}

/*
 * Open audio resources
 */
int slimaudio_open(slimaudio_t *audio) {
	if (slimaudio_http_open(audio) != 0)
		return -1;		
	if (slimaudio_decoder_open(audio) != 0)
		return -1;
	if (slimaudio_output_open(audio) != 0)
		return -1;
		
	return 0;	
}

/*
 * Close audio resources
 */
int slimaudio_close(slimaudio_t *audio) {
	audio_stop(audio);
	
	if (slimaudio_http_close(audio) != 0)
		return -1;
	if (slimaudio_decoder_close(audio) != 0)
		return -1;
	if (slimaudio_output_close(audio) != 0)
		return -1;
		
	return 0;
}

int slimaudio_stat(slimaudio_t *audio, char *code) {
	int decoder_available = slimaudio_buffer_available(audio->decoder_buffer);
	int output_available = slimaudio_buffer_available(audio->output_buffer);
	
	DEBUGF("slimaudio_stat %4.4s %i %i\n", code, decoder_available, output_available);

	return slimproto_stat(audio->proto,
						code,
						DECODER_BUFFER_SIZE,
						decoder_available,
						audio->http_total_bytes,
						OUTPUT_BUFFER_SIZE,
						output_available,
						slimaudio_output_streamtime(audio)
						);
}

/*
 * Callback for 'strm' command.
 */
static int strm_callback(slimproto_t *proto, const unsigned char *buf, int buf_len, void *user_data) {
	slimproto_msg_t msg;
	
	slimaudio_t *audio = (slimaudio_t *) user_data;
	slimproto_parse_command(buf, buf_len, &msg);

	DEBUGF("strm cmd %c\n", msg.strm.command);

	switch (msg.strm.command) {
		case 's': /* start */
			slimaudio_http_connect(audio, &msg);
			slimaudio_decoder_connect(audio, &msg);
			slimaudio_output_connect(audio, &msg);
			break;
			
		case 'p': /* pause */
			slimaudio_output_pause(audio);
			break;	
		
		case 'u': /* unpause */
			slimaudio_output_unpause(audio);
			break;	
		
		case 'q': /* stop */
			audio_stop(audio);
			break;	
		
		case 't': /* status */
			break;			
	}
	
	slimaudio_stat(audio, (char *)&msg.strm.cmd);
	
	DEBUGF("DONE strm cmd %c\n", msg.strm.command);
	return 0;
}


static void audio_stop(slimaudio_t *audio) {
	/*
	 * To prevent deadlocks we must stop the decoder and http 
	 * stream together.
	 */
	slimaudio_output_disconnect(audio);

	pthread_mutex_lock(&audio->decoder_mutex);
	pthread_mutex_lock(&audio->http_mutex);
	
	audio->decoder_state = STREAM_STOP;
	audio->http_state = STREAM_STOP;
	
	pthread_cond_broadcast(&audio->decoder_cond);
	pthread_cond_broadcast(&audio->http_cond);
	
	slimaudio_buffer_flush(audio->output_buffer);
	slimaudio_buffer_flush(audio->decoder_buffer);

	while (audio->decoder_state != STREAM_STOPPED) {
		pthread_cond_wait(&audio->decoder_cond, &audio->decoder_mutex);
	}
	while (audio->http_state != STREAM_STOPPED) {
		pthread_cond_wait(&audio->http_cond, &audio->http_mutex);
	}
	
	pthread_mutex_unlock(&audio->http_mutex);	
	pthread_mutex_unlock(&audio->decoder_mutex);	
}
