/*
 *   SlimProtoLib Copyright (c) 2004,2006 Richard Titmuss
 *
 *   This file is part of SlimProtoLib.
 *
 *   SlimProtoLib is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   SlimProtoLib is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with SlimProtoLib; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <portaudio.h>
#include <portmixer.h>

#include "slimproto/slimproto.h"
#include "slimaudio/slimaudio.h"


#ifdef SLIMPROTO_DEBUG
  #define DEBUGF(...) if (slimaudio_output_debug) fprintf(stderr, __VA_ARGS__)
#else
  #define DEBUGF(...)
#endif


bool slimaudio_output_debug;


static void *output_thread(void *ptr);

static int pa_callback(  void *inputBuffer, void *outputBuffer,
                     unsigned long framesPerBuffer,
                     PaTimestamp outTime, void *userData );


static int audg_callback(slimproto_t *p, const unsigned char *buf, int buf_len, void *user_data);


int slimaudio_output_init(slimaudio_t *audio) {
	int i, err;
	
	err = Pa_Initialize();
	if (err != paNoError) {
		printf("PortAudio error: %s\n", Pa_GetErrorText(err) );	
		exit(-1);
	}	
	DEBUGF("slimaudio_output_init: PortAudio initialised\n");
	
	audio->num_device_names = Pa_CountDevices();
	audio->device_names = (char **)malloc(sizeof(char *) * audio->num_device_names);
	
	const PaDeviceInfo *pdi;
	for(i=0; i<audio->num_device_names; i++ ) {
		pdi = Pa_GetDeviceInfo( i );
		audio->device_names[i] = strdup(pdi->name);
	}

	audio->output_device_id = Pa_GetDefaultOutputDeviceID();
	
	slimproto_add_command_callback(audio->proto, "audg", &audg_callback, audio);
	
	return 0;
}

void slimaudio_output_destroy(slimaudio_t *audio) {
	int i;
	
	// FIXME remove slimproto callback
	
	for (i=0; i<audio->num_device_names; i++) {
		free(audio->device_names[i]);	
	}
	free(audio->device_names);
	
	Pa_Terminate();
}

void slimaudio_get_output_devices(slimaudio_t *audio, char ***device_names, int *num_device_names) {
	*device_names = audio->device_names;
	*num_device_names = audio->num_device_names;
}

void slimaudio_set_output_device(slimaudio_t *audio, int device_id) {
	audio->output_device_id = device_id;
}

int slimaudio_output_open(slimaudio_t *audio) {
	pthread_mutex_init(&(audio->output_mutex), NULL);
	pthread_cond_init(&(audio->output_cond), NULL);
	
	audio->output_state = STOPPED;	

	if (pthread_create( &audio->output_thread, NULL, output_thread, (void*) audio) != 0) {
		fprintf(stderr, "Error creating output thread\n");
		return -1;	
	}	
	return 0;
}


int slimaudio_output_close(slimaudio_t *audio) {
	pthread_mutex_lock(&audio->output_mutex);

	audio->output_state = QUIT;	
	pthread_cond_broadcast(&audio->output_cond);
	
	pthread_mutex_unlock(&audio->output_mutex);
	
	pthread_join(audio->output_thread, NULL);
	return 0;
}

 
static void *output_thread(void *ptr) {
	int err;
	struct timeval  now;
	struct timespec timeout;
	
	slimaudio_t *audio = (slimaudio_t *) ptr;

	err = Pa_OpenStream(
				&audio->pa_stream,		// stream
				paNoDevice,				// input device
				0,						// input channels
				0,						// input sample format
				NULL,					// input driver info
				audio->output_device_id,// output device				
				2,						// output channels
				paInt16,				// output sample format
				NULL,					// output driver info
				44100,					// sample rate
				256,					// frames per buffer
				0,						// number of buffers
				0,
				pa_callback,			// callback
				audio);					// user data
	if (err != paNoError) {
		printf("output_thread: PortAudio error1: %s\n", Pa_GetErrorText(err) );	
		exit(-1);
	}	


	int num_mixers = Px_GetNumMixers(audio->pa_stream);
	while (--num_mixers >= 0) {
		DEBUGF("Mixer: %s\n", Px_GetMixerName(audio->pa_stream, num_mixers));
	}
	
	audio->px_mixer = Px_OpenMixer(audio->pa_stream, 0);
	DEBUGF("Px_mixer = %p\n", audio->px_mixer);

	pthread_mutex_lock(&audio->output_mutex);				

	while (audio->output_state != QUIT) {				
		DEBUGF("output_thread: state %i\n", audio->output_state);

		switch (audio->output_state) {
			case STOPPED:
			case PAUSED:
				pthread_cond_wait(&audio->output_cond, &audio->output_mutex);
				break;

			case PLAY:
/*
 * TODO I am not sure if it's best to start the stream in the play or buffering
 * state ... which has less latency for sync? If the stream is started in 
 * BUFFERING then pause/unpause breaks.
 */
				err = Pa_StartStream(audio->pa_stream);
				if (err != paNoError) {
					printf("output_thread: PortAudio error2: %s\n", Pa_GetErrorText(err) );	
					exit(-1);
				}	

				slimaudio_buffer_set_readopt(audio->output_buffer, BUFFER_NONBLOCKING);
				audio->pa_streamtime_offset = Pa_StreamTime(audio->pa_stream);				

				audio->output_state = PLAYING;
				pthread_cond_broadcast(&audio->output_cond);
				break;
				
			case BUFFERING:
			case PLAYING:			
				gettimeofday(&now, NULL);
				timeout.tv_sec = now.tv_sec + 1;
				timeout.tv_nsec = now.tv_usec * 1000;				
				err = pthread_cond_timedwait(&audio->output_cond, &audio->output_mutex, &timeout);
					
				if (err == ETIMEDOUT) {
					slimaudio_stat(audio, "STMt");
				}
					
				if (audio->output_STMu) {
					audio->output_STMu = false;
					slimaudio_stat(audio, "STMu");						
				}
					
				if (audio->output_STMs) {
					audio->output_STMs = false;
					audio->pa_streamtime_offset = Pa_StreamTime(audio->pa_stream);				
					slimaudio_stat(audio, "STMs");						
				}
				break;
		
			case STOP:
				err = Pa_AbortStream(audio->pa_stream);
				if (err != paNoError) {
					printf("output_thread: PortAudio error3: %s\n", Pa_GetErrorText(err) );	
					exit(-1);
				}
					
				audio->output_state = STOPPED;
				pthread_cond_broadcast(&audio->output_cond);
				break;
				
			case PAUSE:
				err = Pa_StopStream(audio->pa_stream);
				if (err != paNoError) {
					printf("output_thread: PortAudio error3: %s\n", Pa_GetErrorText(err) );	
					exit(-1);
				}					
					
				audio->output_state = PAUSED;	
				pthread_cond_broadcast(&audio->output_cond);
				break;

			case QUIT:
				break;
		}		
	}
	pthread_mutex_unlock(&audio->output_mutex);
	
	Px_CloseMixer(audio->px_mixer);
	audio->px_mixer = NULL;
	
	err = Pa_CloseStream(audio->pa_stream);
	if (err != paNoError) {
		printf("output_thread: PortAudio error3: %s\n", Pa_GetErrorText(err) );	
		exit(-1);
	}
	audio->pa_stream = NULL;

	return 0;
}


/*
 * Callback for the 'audg' command
 */
static int audg_callback(slimproto_t *proto, const unsigned char *buf, int buf_len, void *user_data) {
	slimproto_msg_t msg;
	
	slimaudio_t *audio = (slimaudio_t *) user_data;
	slimproto_parse_command(buf, buf_len, &msg);

	PxVolume volume = (float)msg.audg.left_gain / (float)0x10000;
	DEBUGF("audg cmd %04x %f\n", msg.audg.left_gain, volume);

	if (audio->px_mixer != NULL) {
		Px_SetPCMOutputVolume(audio->px_mixer, volume);
		DEBUGF("pcm volume %f\n", Px_GetPCMOutputVolume(audio->px_mixer));	
	}
		
	return 0;
}


void slimaudio_output_connect(slimaudio_t *audio, slimproto_msg_t *msg) {
	pthread_mutex_lock(&audio->output_mutex);

	if (audio->output_state == PLAYING) {
		pthread_mutex_unlock(&audio->output_mutex);
		return;
	}
	
	DEBUGF("slimaudio_output_connect: state=%i\n", audio->output_state);
	audio->output_state = BUFFERING;
	pthread_cond_broadcast(&audio->output_cond);

	pthread_mutex_unlock(&audio->output_mutex);	
}


void slimaudio_output_disconnect(slimaudio_t *audio) {
	pthread_mutex_lock(&audio->output_mutex);

	DEBUGF("slimaudio_output_disconnect: state=%i\n", audio->output_state);

	if (audio->output_state == STOPPED) {
		pthread_mutex_unlock(&audio->output_mutex);	
		return;	
	}	

	audio->output_state = STOP;	
	pthread_cond_broadcast(&audio->output_cond);

	while (audio->output_state == STOP) {				
		pthread_cond_wait(&audio->output_cond, &audio->output_mutex);
	}

	pthread_mutex_unlock(&audio->output_mutex);	
}


void slimaudio_output_pause(slimaudio_t *audio) {
	pthread_mutex_lock(&audio->output_mutex);
	
	DEBUGF("slimaudio_output_pause: state=%i\n", audio->output_state);

	audio->output_state = PAUSE;	
	pthread_cond_broadcast(&audio->output_cond);

	while (audio->output_state == PAUSE) {				
		pthread_cond_wait(&audio->output_cond, &audio->output_mutex);
	}

	pthread_mutex_unlock(&audio->output_mutex);	
}


void slimaudio_output_unpause(slimaudio_t *audio) {
	pthread_mutex_lock(&audio->output_mutex);

	if (audio->output_state == PLAYING) {
		pthread_mutex_unlock(&audio->output_mutex);	
		return;	
	}

	DEBUGF("slimaudio_output_unpause: state=%i\n", audio->output_state);

	audio->output_state = PLAY;
	pthread_cond_broadcast(&audio->output_cond);

	while (audio->output_state == PLAY) {
		pthread_cond_wait(&audio->output_cond, &audio->output_mutex);
	}


	pthread_mutex_unlock(&audio->output_mutex);	
}


int slimaudio_output_streamtime(slimaudio_t *audio) {
	if (audio->output_state != PLAYING)
		return 0;
		
	PaTimestamp numSamples = Pa_StreamTime(audio->pa_stream);
	return (int) ((numSamples - audio->pa_streamtime_offset) / 44100.0);	
}



static int pa_callback(  void *inputBuffer, void *outputBuffer,
                     unsigned long framesPerBuffer,
                     PaTimestamp outTime, void *userData )
{
	slimaudio_t *audio = (slimaudio_t *) userData;

	int len = framesPerBuffer * 2 * 2; // 2 channels, 16 bit samples
	
	int off = 0;
	while (audio->output_state == PLAYING && (len - off) > 0 ) {
		slimaudio_buffer_status ok;
			
		int data_len = len-off;
		ok = slimaudio_buffer_read(audio->output_buffer, (char *) outputBuffer+off, &data_len);
		if (ok == SLIMAUDIO_BUFFER_STREAM_END) {
			/* stream closed */
				
			if (slimaudio_buffer_available(audio->output_buffer) == 0) {
				pthread_mutex_lock(&audio->output_mutex);

				/*
				 * Send buffer underrun to slimserver. During normal play this indicates
				 * the end of the playlist. During sync this starts the next track. 
				 */
				audio->output_STMu = true;

				pthread_cond_broadcast(&audio->output_cond);
				pthread_mutex_unlock(&audio->output_mutex);
			}
		}
		else if (ok == SLIMAUDIO_BUFFER_STREAM_START) {
			pthread_mutex_lock(&audio->output_mutex);
				
			/*
			 * Send track start to slimserver. During normal play this advances the
			 * playlist.
			 */
			audio->output_STMs = true;

			pthread_cond_broadcast(&audio->output_cond);
			pthread_mutex_unlock(&audio->output_mutex);
		}

		off += data_len;

		/* if we have underrun fill remaining buffer with silence */
		if (data_len == 0) {
			memset((char *)outputBuffer+off, 0, len-off);
			off = len;
		}
	}
	
//	printf("pa_callback complete framePerBuffer=%i n=%i\n", framesPerBuffer, n);
	
	return 0;
}


